# -*- coding: utf-8 -*-

import atexit

from .celery_executor import CeleryExecutor
from .sequential_executor import SequentialExecutor
from .. import configuration as conf


DEFAULT_EXECUTOR = None


class Executors:
    CeleryExecutor = 'CeleryExecutor'
    SequentialExecutor = 'SequentialExecutor'


def get_default_executor():
    global DEFAULT_EXECUTOR

    if DEFAULT_EXECUTOR is not None:
        return DEFAULT_EXECUTOR

    configure_executor(conf.get('core', 'EXECUTOR'))

    return DEFAULT_EXECUTOR


def configure_executor(executor_name):
    global DEFAULT_EXECUTOR

    if DEFAULT_EXECUTOR is None:
        if executor_name == Executors.CeleryExecutor:
            DEFAULT_EXECUTOR = CeleryExecutor()
        elif executor_name == Executors.SequentialExecutor:
            DEFAULT_EXECUTOR = SequentialExecutor()
        else:
            raise NotImplementedError('Unknown executor')


def dispose_executor():
    global DEFAULT_EXECUTOR

    if DEFAULT_EXECUTOR is not None:
        DEFAULT_EXECUTOR.stop()


configure_executor(conf.get('core', 'EXECUTOR'))
atexit.register(dispose_executor)



# -*- coding: utf-8 -*-

from .base_operator import BaseOperator


class PythonOperator(BaseOperator):
    def __init__(self, python_callable, op_args=None, op_kwargs=None):
        self.python_callable = python_callable
        self.op_args = op_args or []
        self.op_kwargs = op_kwargs or {}

    def execute(self):
        self.execute_callable()

    def execute_callable(self):
        self.python_callable(*self.op_args, **self.op_kwargs)



# -*- coding: utf-8 -*-

import logging
import traceback
import subprocess

from celery import Celery

from .base_executor import BaseExecutor
from .. import configuration as conf
from ..config_templates.default_celery import DEFAULT_CELERY_CONFIG
from ..utils.module_loading import import_string


if conf.has_option('celery', 'celery_config_options'):
    celery_configuration = import_string(
        conf.get('celery', 'celery_config_options'))
else:
    celery_configuration = DEFAULT_CELERY_CONFIG


app = Celery(
    conf.get('celery', 'CELERY_APP_NAME'),
    config_source=celery_configuration,
    fixups=[])


@app.task
def execute_command(command):
    try:
        subprocess.check_call(command, stderr=subprocess.STDOUT, close_fds=True)
    except Exception as e:
        traceback.print_exc()


class CeleryExecutor(BaseExecutor):
    def __init__(self):
        pass

    def start(self):
        pass

    def execute_async(self, command):
        execute_command.apply_async(args=[command])

    def stop(self):
        pass



# https://github.com/apache/incubator-airflow/blob/master/airflow/utils/module_loading.py
# https://github.com/apache/incubator-airflow/blob/master/LICENSE
from importlib import import_module


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        raise ImportError("{} doesn't look like a module path".format(dotted_path))

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError('Module "{}" does not define a "{}" attribute/class'.format(
            module_path, class_name)
        )



# -*- coding: utf-8 -*-

import os
import json
import datetime as dt

from sqlalchemy import Boolean, Column, DateTime, Integer, String, Text, ForeignKey
from sqlalchemy.ext.declarative import declarative_base


Base = declarative_base()
ID_LEN = 100
STR_LEN = 500


class File(Base):
    __tablename__ = "files"

    file_id = Column(String(ID_LEN), primary_key=True)
    uploaded_at = Column(DateTime)
    page_number = Column(Integer)
    filename = Column(String(STR_LEN))
    filepath = Column(String(STR_LEN))
    has_image = Column(Boolean, default=False)
    imagename = Column(String(STR_LEN))
    imagepath = Column(String(STR_LEN))
    file_dimensions = Column(Text)
    image_dimensions = Column(Text)
    detected_areas = Column(Text)


class Rule(Base):
    __tablename__ = "rules"

    rule_id = Column(String(ID_LEN), primary_key=True)
    created_at = Column(DateTime)
    rule_name = Column(String(STR_LEN))
    rule_options = Column(Text)


class Job(Base):
    __tablename__ = "jobs"

    job_id = Column(String(ID_LEN), primary_key=True)
    page_numbers = Column(Text)
    datapath = Column(String(STR_LEN), default=None)
    render_files = Column(Text, default=json.dumps([]))
    is_finished = Column(Boolean, default=False)
    started_at = Column(DateTime)
    finished_at = Column(DateTime, default=dt.datetime.now())
    file_id = Column(String(ID_LEN), ForeignKey('files.file_id'))
    rule_id = Column(String(ID_LEN), ForeignKey('rules.rule_id'))



from .. import configuration as conf


broker_url = conf.get('celery', 'BROKER_URL')

DEFAULT_CELERY_CONFIG = {
    'broker_url': broker_url,
    'worker_prefetch_multiplier': 1,
    'task_acks_late': True,
    'task_create_missing_queues': True,
    'task_default_queue': conf.get('celery', 'DEFAULT_QUEUE'),
    'worker_concurrency': int(conf.get('celery', 'WORKER_CONCURRENCY'))
}



import uuid
import random
import string


def generate_uuid():
    return str(uuid.uuid4())


def random_string(length):
    ret = ''
    while length:
        ret += random.choice(string.digits + string.ascii_lowercase + string.ascii_uppercase)
        length -= 1
    return ret



# -*- coding: utf-8 -*-

import os
import six
import shutil

from backports.configparser import ConfigParser


def _read_default_config_file(file_name):
    """
    https://github.com/apache/incubator-airflow/blob/master/airflow/configuration.py
    https://github.com/apache/incubator-airflow/blob/master/LICENSE
    """
    templates_dir = os.path.join(os.path.dirname(__file__), 'config_templates')
    file_path = os.path.join(templates_dir, file_name)
    if six.PY2:
        with open(file_path) as f:
            config = f.read()
            return config.decode('utf-8')
    else:
        with open(file_path, encoding='utf-8') as f:
            return f.read()


def expand_env_var(env_var):
    """
    https://github.com/apache/incubator-airflow/blob/master/airflow/configuration.py
    https://github.com/apache/incubator-airflow/blob/master/LICENSE

    Expands (potentially nested) env vars by repeatedly applying
    `expandvars` and `expanduser` until interpolation stops having
    any effect.
    """
    if not env_var:
        return env_var
    while True:
        interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
        if interpolated == env_var:
            return interpolated
        else:
            env_var = interpolated


DEFAULT_CONFIG = _read_default_config_file('default_excalibur.cfg')


class ExcaliburConfigParser(ConfigParser):
    def __init__(self, default_config=None, *args, **kwargs):
        super(ExcaliburConfigParser, self).__init__(*args, **kwargs)

        self.excalibur_defaults = ConfigParser(*args, **kwargs)
        if default_config is not None:
            self.excalibur_defaults.read_string(default_config)

        self.is_validated = False

    def _validate(self):
        if self.get("core", "executor") != 'SequentialExecutor' and \
                "sqlite" in self.get('core', 'sql_alchemy_conn'):
            raise ValueError('Cannot use sqlite with the {}'.format(
                self.get('core', 'executor')))

        self.is_validated = True

    def get(self, section, key, **kwargs):
        section = str(section).lower()
        key = str(key).lower()

        if super(ExcaliburConfigParser, self).has_option(section, key):
            return expand_env_var(
                super(ExcaliburConfigParser, self).get(section, key, **kwargs))

        if self.excalibur_defaults.has_option(section, key):
            return expand_env_var(
                self.excalibur_defaults.get(section, key, **kwargs))

        else:
            raise ValueError('section/key [{section}/{key}] not found in'
                             ' config'.format(**locals()))

    def read(self, filename):
        super(ExcaliburConfigParser, self).read(filename)
        self._validate()


def mkdirs(path):
    if not os.path.isdir(path):
        os.makedirs(path)


if 'EXCALIBUR_HOME' not in os.environ:
    EXCALIBUR_HOME = expand_env_var('~/excalibur')
else:
    EXCALIBUR_HOME = expand_env_var(os.environ['EXCALIBUR_HOME'])

mkdirs(EXCALIBUR_HOME)

if 'EXCALIBUR_CONFIG' not in os.environ:
    EXCALIBUR_CONFIG = EXCALIBUR_HOME + '/excalibur.cfg'
else:
    EXCALIBUR_CONFIG = expand_env_var(os.environ['EXCALIBUR_CONFIG'])


def parameterized_config(template):
    """
    https://github.com/apache/incubator-airflow/blob/master/airflow/configuration.py
    https://github.com/apache/incubator-airflow/blob/master/LICENSE

    Generates a configuration from the provided template + variables defined in
    current scope
    :param template: a config content templated with {{variables}}
    """
    all_vars = {k: v for d in [globals(), locals()] for k, v in d.items()}
    return template.format(**all_vars)


if not os.path.isfile(EXCALIBUR_CONFIG):
    print('Creating new Excalibur configuration file in: {}'.format(
        EXCALIBUR_CONFIG))
    with open(EXCALIBUR_CONFIG, 'w') as f:
        cfg = parameterized_config(DEFAULT_CONFIG)
        if six.PY2:
            cfg = cfg.encode('utf8')
        f.write(cfg)


conf = ExcaliburConfigParser(default_config=parameterized_config(DEFAULT_CONFIG))

conf.read(EXCALIBUR_CONFIG)

# for Flask
ALLOWED_EXTENSIONS = ['pdf', 'json']
SECRET_KEY = conf.get('webserver', 'SECRET_KEY')
PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))
PDFS_FOLDER = os.path.join(PROJECT_ROOT, 'www/static/uploads')
USING_SQLITE = True if conf.get('core', 'SQL_ALCHEMY_CONN').startswith('sqlite') else False

get = conf.get
has_option = conf.has_option



# -*- coding: utf-8 -*-

import os
import atexit

from sqlalchemy import create_engine, exc
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.pool import NullPool

from . import configuration as conf


EXCALIBUR_HOME = None
SQL_ALCHEMY_CONN = None

engine = None
Session = None


def configure_vars():
    global EXCALIBUR_HOME
    global SQL_ALCHEMY_CONN
    EXCALIBUR_HOME = os.path.expanduser(conf.get('core', 'EXCALIBUR_HOME'))
    SQL_ALCHEMY_CONN = conf.get('core', 'SQL_ALCHEMY_CONN')


def configure_orm():
    global engine
    global Session

    engine_args = {
        'poolclass': NullPool
    }
    engine = create_engine(SQL_ALCHEMY_CONN, **engine_args)
    Session = scoped_session(sessionmaker(
        autocommit=False, autoflush=False, bind=engine, expire_on_commit=False))


def dispose_orm():
    global engine
    global Session

    if Session is not None:
        Session.remove()
        Session = None
    if engine is not None:
        engine.dispose()
        engine = None


configure_vars()
configure_orm()
atexit.register(dispose_orm)



# -*- coding: utf-8 -*-

class BaseExecutor(object):
    def __init__(self):
        pass

    def start(self):
        raise NotImplementedError()

    def execute_async(self, task, *args, **kwargs):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()



# -*- coding: utf-8 -*-

import os
import glob
import json
import logging
import subprocess
import datetime as dt

import camelot

from . import configuration as conf
from .models import File, Rule, Job
from .settings import Session
from .utils.file import mkdirs
from .utils.task import (save_page, get_page_layout, get_file_dimensions,
                         get_image_dimensions)


def split(file_id):
    try:
        def get_executable():
            import platform
            from distutils.spawn import find_executable

            class GhostscriptNotFound(Exception): pass

            gs = None
            system = platform.system().lower()
            try:
                if system == 'windows':
                    if find_executable('gswin32c.exe'):
                        gs = 'gswin32c.exe'
                    elif find_executable('gswin64c.exe'):
                        gs = 'gswin64c.exe'
                    else:
                        raise ValueError
                else:
                    if find_executable('gs'):
                        gs = 'gs'
                    elif find_executable('gsc'):
                        gs = 'gsc'
                    else:
                        raise ValueError
                if 'ghostscript' not in subprocess.check_output(
                        [gs, '-version']).decode('utf-8').lower():
                    raise ValueError
            except ValueError:
                raise GhostscriptNotFound(
                    'Please make sure that Ghostscript is installed'
                    ' and available on the PATH environment variable')

            return gs

        session = Session()
        file = session.query(File).filter(File.file_id == file_id).first()
        save_page(file.filepath, file.page_number)

        filename = 'page-{}.pdf'.format(file.page_number)
        filepath = os.path.join(conf.PDFS_FOLDER, file_id, filename)
        imagename = ''.join([filename.replace('.pdf', ''), '.png'])
        imagepath = os.path.join(conf.PDFS_FOLDER, file_id, imagename)

        gs_call = [
            '-q',
            '-sDEVICE=png16m',
            '-o',
            imagepath,
            '-r600',
            filepath
        ]
        gs = get_executable()
        gs_call.insert(0, gs)
        process = subprocess.Popen(gs_call)
        out = process.communicate()[0]
        ret = process.wait()

        file.has_image = True
        file.imagename = imagename
        file.imagepath = imagepath

        file_dimensions = get_file_dimensions(filepath)
        image_dimensions = get_image_dimensions(imagepath)
        pdf_width_scaler = image_dimensions[0] / float(file_dimensions[0])
        pdf_height_scaler = image_dimensions[1] / float(file_dimensions[1])

        file.file_dimensions = json.dumps(file_dimensions)
        file.image_dimensions = json.dumps(image_dimensions)

        lattice_areas, stream_areas = [None] * 2
        # lattice
        tables = camelot.read_pdf(filepath, flavor='lattice')
        if len(tables):
            lattice_areas = []
            for table in tables:
                x1, y1, x2, y2 = tables[0]._bbox
                lattice_areas.append((x1, y2, x2, y1))
        # stream
        tables = camelot.read_pdf(filepath, flavor='stream')
        if len(tables):
            stream_areas = []
            for table in tables:
                x1, y1, x2, y2 = tables[0]._bbox
                stream_areas.append((x1, y2, x2, y1))

        detected_areas = {
            'lattice': lattice_areas,
            'stream': stream_areas
        }
        file.detected_areas = json.dumps(detected_areas)

        session.commit()
        session.close()
    except Exception as e:
        logging.exception(e)


def extract(job_id):
    try:
        session = Session()
        job = session.query(Job).filter(Job.job_id == job_id).first()
        rule = session.query(Rule).filter(Rule.rule_id == job.rule_id).first()
        file = session.query(File).filter(File.file_id == job.file_id).first()

        rule_options = json.loads(rule.rule_options)
        flavor = rule_options.pop('flavor')
        tables = camelot.read_pdf(file.filepath, pages=job.page_numbers, flavor=flavor.lower(), **rule_options)

        froot, fext = os.path.splitext(file.filename)
        datapath = os.path.dirname(file.filepath)
        for f in ['csv', 'excel', 'json', 'html']:
            f_datapath = os.path.join(datapath, f)
            mkdirs(f_datapath)
            ext = f if f != 'excel' else 'xlsx'
            f_datapath = os.path.join(f_datapath, '{}.{}'.format(froot, ext))
            tables.export(f_datapath, f=f, compress=True)

        # for render
        jsonpath = os.path.join(datapath, 'json')
        jsonpath = os.path.join(jsonpath, '{}.json'.format(froot))
        tables.export(jsonpath, f='json')
        render_files = {os.path.splitext(os.path.basename(f))[0]: f
            for f in glob.glob(os.path.join(datapath, 'json/*.json'))}

        job.datapath = datapath
        job.render_files = json.dumps(render_files)
        job.is_finished = True
        job.finished_at = dt.datetime.now()

        session.commit()
        session.close()
    except Exception as e:
        logging.exception(e)



import json

from flask import Flask, Blueprint
from werkzeug.utils import find_modules, import_string

from .. import configuration as conf
from .views import views


def to_pretty_json(value):
    value = json.loads(value)
    return json.dumps(value, sort_keys=True,
                      indent=4, separators=(',', ': '))


def create_app(config=None):
    app = Flask(__name__)
    app.config.from_object(conf)
    app.register_blueprint(views)
    app.jinja_env.filters['pretty'] = to_pretty_json
    return app



from ..settings import engine


def initialize_database():
    from ..models import Base

    Base.metadata.create_all(engine)


def reset_database():
    from ..models import Base

    Base.metadata.drop_all(engine)
    initialize_database()


import os

import cv2
from PyPDF2 import PdfFileReader, PdfFileWriter
from camelot.utils import get_page_layout, get_text_objects, get_rotation


def save_page(filepath, page_number):
    infile = PdfFileReader(open(filepath, 'rb'), strict=False)
    page = infile.getPage(page_number - 1)
    outfile = PdfFileWriter()
    outfile.addPage(page)
    outpath = os.path.join(os.path.dirname(filepath), 'page-{}.pdf'.format(page_number))
    with open(outpath, 'wb') as f:
        outfile.write(f)
    froot, fext = os.path.splitext(outpath)
    layout, __ = get_page_layout(outpath)
    # fix rotated PDF
    lttextlh = get_text_objects(layout, ltype="lh")
    lttextlv = get_text_objects(layout, ltype="lv")
    ltchar = get_text_objects(layout, ltype="char")
    rotation = get_rotation(lttextlh, lttextlv, ltchar)
    if rotation != '':
        outpath_new = ''.join([froot.replace('page', 'p'), '_rotated', fext])
        os.rename(outpath, outpath_new)
        infile = PdfFileReader(open(outpath_new, 'rb'), strict=False)
        if infile.isEncrypted:
            infile.decrypt('')
        outfile = PdfFileWriter()
        p = infile.getPage(0)
        if rotation == 'anticlockwise':
            p.rotateClockwise(90)
        elif rotation == 'clockwise':
            p.rotateCounterClockwise(90)
        outfile.addPage(p)
        with open(outpath, 'wb') as f:
            outfile.write(f)


def get_file_dimensions(filepath):
    layout, dimensions = get_page_layout(filepath)
    return list(dimensions)


def get_image_dimensions(imagepath):
    image = cv2.imread(imagepath)
    return [image.shape[1], image.shape[0]]



import os

from .. import configuration as conf


def mkdirs(path):
    if not os.path.isdir(path):
        os.makedirs(path)


def allowed_filename(filename):
    return '.' in filename and filename.rsplit('.', 1)[1].lower() in conf.ALLOWED_EXTENSIONS



# -*- coding: utf-8 -*-

import traceback
import subprocess
from concurrent.futures import ProcessPoolExecutor

from .base_executor import BaseExecutor


def execute_command(command):
    try:
        subprocess.check_call(command, stderr=subprocess.STDOUT, close_fds=True)
    except FileNotFoundError:
        # TODO: PyInstaller does not package console_scripts
        # https://github.com/pyinstaller/pyinstaller/issues/305
        from ..cli import _run

        task_name = command[-3]
        task_id = command[-1]
        _run(task_name, task_id)
    except Exception as e:
        traceback.print_exc()


class SequentialExecutor(BaseExecutor):
    def __init__(self):
        self.start()

    def start(self):
        self.pool = ProcessPoolExecutor(1)

    def execute_async(self, command):
        self.pool.submit(execute_command, command)

    def stop(self):
        self.pool.shutdown(wait=True)



# -*- coding: utf-8 -*-

import os

import click

from . import __version__, settings
from . import configuration as conf
from .operators.python_operator import PythonOperator
from .tasks import split, extract
from .utils.database import initialize_database, reset_database
from .www.app import create_app


def abort_if_false(ctx, param, value):
    if not value:
        ctx.abort()


def _run(task_name, task_id):
    task_bag = {
        'split': split,
        'extract': extract
    }
    python_callable = task_bag[task_name]
    task = PythonOperator(python_callable, op_args=[task_id])
    task.execute()


@click.group()
@click.version_option(version=__version__)
def cli(*args, **kwargs):
    pass


@cli.command('initdb')
def initdb(*args, **kwargs):
    initialize_database()


@cli.command('resetdb')
def resetdb(*args, **kwargs):
    click.confirm(
        "This will drop existing tables if they exist. Proceed?", abort=True)

    reset_database()
    initialize_database()


@cli.command('webserver')
def webserver(*args, **kwargs):
    if conf.USING_SQLITE:
        sqlite_path = settings.SQL_ALCHEMY_CONN.replace('sqlite:///', '')
        if not os.path.isfile(sqlite_path):
            initialize_database()

    app = create_app(conf)
    app.run(use_reloader=False)


@cli.command('worker')
def worker(*args, **kwargs):
    from celery.bin import worker
    from .executors.celery_executor import app as celery_app


    worker = worker.worker(app=celery_app)
    options = {
        'concurrency': int(conf.get('celery', 'WORKER_CONCURRENCY')),
        'loglevel': conf.get('core', 'LOGGING_LEVEL')
    }
    worker.run(**options)


@cli.command('run')
@click.option('-t', '--task')
@click.option('-id', '--uuid')
def run(*args, **kwargs):
    _run(kwargs['task'], kwargs['uuid'])



# -*- coding: utf-8 -*-

import os
import re
import glob
import json
import datetime as dt

import pandas as pd
from werkzeug import secure_filename
from flask import (Blueprint, jsonify, redirect, render_template, request,
                   send_from_directory, url_for)

from .. import configuration as conf
from ..executors import get_default_executor
from ..models import File, Rule, Job
from ..settings import Session
from ..utils.file import mkdirs, allowed_filename
from ..utils.metadata import generate_uuid, random_string


views = Blueprint('views', __name__)


@views.route('/', methods=['GET'])
def index():
    return redirect(url_for('views.files'))


@views.route('/files', methods=['GET', 'POST'])
def files():
    if request.method == 'GET':
        files_response = []
        session = Session()
        for file in session.query(File).order_by(File.uploaded_at.desc()).all():
            job = session.query(Job).filter(Job.file_id == file.file_id).order_by(Job.started_at.desc()).first()
            files_response.append({
                'file_id': file.file_id,
                'job_id': job.job_id if job is not None else '',
                'uploaded_at': file.uploaded_at.strftime('%Y-%m-%dT%H:%M:%S'),
                'filename': file.filename
            })
        session.close()
        return render_template('files.html', files_response=files_response)
    file = request.files['file-0']
    if file and allowed_filename(file.filename):
        file_id = generate_uuid()
        uploaded_at = dt.datetime.now()
        page_number = int(request.form['page_number'])
        filename = secure_filename(file.filename)
        filepath = os.path.join(conf.PDFS_FOLDER, file_id)
        mkdirs(filepath)
        filepath = os.path.join(filepath, filename)
        file.save(filepath)

        session = Session()
        f = File(
            file_id=file_id,
            uploaded_at=uploaded_at,
            page_number=page_number,
            filename=filename,
            filepath=filepath
        )
        session.add(f)
        session.commit()
        session.close()

        command = 'excalibur run --task {} --uuid {}'.format('split', file_id)
        command_as_list = command.split(' ')
        executor = get_default_executor()
        executor.execute_async(command_as_list)
    return jsonify(file_id=file_id)


@views.route('/workspaces/<string:file_id>', methods=['GET'])
def workspaces(file_id):
    session = Session()
    file = session.query(File).filter(File.file_id == file_id).first()
    rules = session.query(Rule).all()
    session.close()
    imagepath, file_dimensions, image_dimensions, detected_areas, saved_rules = [None] * 5
    if file.has_image:
        imagepath = file.imagepath.replace(
            os.path.join(conf.PROJECT_ROOT, 'www'), '')
        file_dimensions = json.loads(file.file_dimensions)
        image_dimensions = json.loads(file.image_dimensions)
        detected_areas = json.loads(file.detected_areas)
        saved_rules = [
            {
                'rule_id': rule.rule_id,
                'rule_name': rule.rule_name
            }
            for rule in rules]
    return render_template(
        'workspace.html', imagepath=imagepath, file_dimensions=file_dimensions,
        image_dimensions=image_dimensions, detected_areas=detected_areas,
        saved_rules=saved_rules)


@views.route('/rules', methods=['GET', 'POST'], defaults={'rule_id': None})
@views.route('/rules/<string:rule_id>', methods=['GET'])
def rules(rule_id):
    if request.method == 'GET':
        if rule_id is not None:
            session = Session()
            rule = session.query(Rule).filter(Rule.rule_id == rule_id).first()
            session.close()
            message = 'Rule not found'
            rule_options = {}
            if rule is not None:
                message = ''
                rule_options = json.loads(rule.rule_options)
            return jsonify(message=message, rule_options=rule_options)
        session = Session()
        rules = session.query(Rule).order_by(Rule.created_at.desc()).all()
        session.close()
        saved_rules = [
            {
                'rule_id': rule.rule_id,
                'created_at': rule.created_at.strftime('%Y-%m-%dT%H:%M:%S'),
                'rule_name': rule.rule_name,
                'rule_options': rule.rule_options
            }
            for rule in rules]
        return render_template('rules.html', saved_rules=saved_rules)
    message='Rule invalid'
    file = request.files['file-0']
    if file and allowed_filename(file.filename):
        rule_id = generate_uuid()
        created_at = dt.datetime.now()
        rule_name = os.path.splitext(secure_filename(file.filename))[0]
        rule_options = file.read()
        message = 'Rule saved'

        session = Session()
        r = Rule(
            rule_id=rule_id,
            created_at=created_at,
            rule_name=rule_name,
            rule_options=rule_options
        )
        session.add(r)
        session.commit()
        session.close()
    return jsonify(message=message)


@views.route('/jobs', methods=['GET', 'POST'], defaults={'job_id': None})
@views.route('/jobs/<string:job_id>', methods=['GET'])
def jobs(job_id):
    if request.method == 'GET':
        if job_id is not None:
            session = Session()
            job = session.query(Job).filter(Job.job_id == job_id).first()
            session.close()

            data = []
            render_files = json.loads(job.render_files)
            regex = 'page-(\d)+-table-(\d)+'
            for k in sorted(render_files,
                    key=lambda x: (int(re.split(regex, x)[1]), int(re.split(regex, x)[2]))):
                df = pd.read_json(render_files[k])
                columns = df.columns.values
                records = df.to_dict('records')
                data.append({
                    'title': k,
                    'columns': columns,
                    'records': records
                })
            return render_template('job.html', is_finished=job.is_finished,
                started_at=job.started_at, finished_at=job.finished_at,
                datapath=job.datapath, data=data)
        return render_template('jobs.html')
    file_id = request.form['file_id']
    rule_id = request.form['rule_id']

    session = Session()
    file = session.query(File).filter(File.file_id == file_id).first()
    session.close()

    if not rule_id:
        rule_id = generate_uuid()
        created_at = dt.datetime.now()
        rule_name = '_'.join([os.path.splitext(file.filename)[0], random_string(6)])
        rule_options = request.form['rule_options']

        session = Session()
        r = Rule(
            rule_id=rule_id,
            created_at=created_at,
            rule_name=rule_name,
            rule_options=rule_options
        )
        session.add(r)
        session.commit()
        session.close()

    job_id = generate_uuid()
    started_at = dt.datetime.now()

    page_numbers = request.form['page_numbers']
    if page_numbers == '1':
        page_numbers = file.page_number

    session = Session()
    j = Job(
        job_id=job_id,
        started_at=started_at,
        page_numbers=page_numbers,
        file_id=file_id,
        rule_id=rule_id
    )
    session.add(j)
    session.commit()
    session.close()

    command = 'excalibur run --task {} --uuid {}'.format('extract', job_id)
    command_as_list = command.split(' ')
    executor = get_default_executor()
    executor.execute_async(command_as_list)
    return jsonify(job_id=job_id)


@views.route('/download', methods=['POST'])
def download():
    job_id = request.form['job_id']
    f = request.form['format']

    session = Session()
    job = session.query(Job).filter(Job.job_id == job_id).first()
    session.close()

    datapath = os.path.join(job.datapath, f.lower())
    zipfile = glob.glob(os.path.join(datapath, '*.zip'))[0]

    directory = os.path.join(os.getcwd(), datapath)
    filename = os.path.basename(zipfile)
    return send_from_directory(
        directory=directory, filename=filename, as_attachment=True)
